/*
 * Copyright (C) 2005-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.notification.listener;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;

import java.util.ArrayList;
import java.util.Enumeration;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jibx.runtime.BindingDirectory;
import org.jibx.runtime.IBindingFactory;
import org.jibx.runtime.IUnmarshallingContext;
import org.jibx.runtime.JiBXException;

import com.funambol.framework.core.Alert;
import com.funambol.framework.core.Constants;
import com.funambol.framework.core.Cred;
import com.funambol.framework.core.Item;
import com.funambol.framework.core.SyncML;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;
import com.funambol.framework.logging.Sync4jLoggerName;
import com.funambol.framework.notification.Message;
import com.funambol.framework.notification.builder.DSNotificationMessageBuilder;
import com.funambol.framework.protocol.ProtocolException;
import com.funambol.framework.security.Officer;
import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.error.ServerException;
import com.funambol.framework.server.inventory.DeviceInventory;
import com.funambol.framework.server.inventory.DeviceInventoryException;
import com.funambol.framework.tools.IOTools;
import com.funambol.framework.tools.SyncMLUtil;
import com.funambol.framework.tools.WBXMLTools;

import com.funambol.server.config.Configuration;
import com.funambol.server.notification.builder.DSNotificationMessageBuilderImpl;

/**
 * Role of this servlet is to receive an address notification message,
 * authenticate the device and if the authentication is successful, register
 * the given address in the DeviceInventory
 *
 * @see com.funambol.framework.server.DeviceInventory
 *
 * @version $Id: Sync4jAddressListener.java,v 1.12 2007/02/03 16:39:49 nichele Exp $
 */
public class Sync4jAddressListener extends HttpServlet {

    // ------------------------------------------------------------ Private data
    private FunambolLogger log = FunambolLoggerFactory.getLogger(Sync4jLoggerName.SERVER_NOTIFICATION);

    private static final String SOURCEURI_IP     = "ip"    ;
    private static final String SOURCEURI_MSISDN = "msisdn";

    private static final int SIZE_INPUT_BUFFER   = 1024;

    private static final String PARAM_DEFAULT_NOTIFICATION_BUILDER =
            "default-notification-builder";

    private static final String PARAM_DEFAULT_TCPIP_NOTIFICATION_SENDER =
            "default-tcpip-notification-sender";

    private static final String PARAM_DEFAULT_WAP_NOTIFICATION_SENDER =
            "default-wap-notification-sender";

    private String defaultNotificationBuilder      = null;
    private String defaultTCPIPNotificationSender  = null;
    private String defaultWAPNotificationSender    = null;

    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {

        ServletConfig config = getServletConfig();

        defaultNotificationBuilder =
            config.getInitParameter(PARAM_DEFAULT_NOTIFICATION_BUILDER);

        defaultTCPIPNotificationSender  =
            config.getInitParameter(PARAM_DEFAULT_TCPIP_NOTIFICATION_SENDER);

        defaultWAPNotificationSender  =
            config.getInitParameter(PARAM_DEFAULT_WAP_NOTIFICATION_SENDER);
    }


    public void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

        if (log.isTraceEnabled()) {
            showHeaders(request);
        }

        if (log.isInfoEnabled()) {
            log.info("Handling incoming request for address listener.");
        }

        //
        //Read the message
        //
        byte[] binaryData = null;
        String xmlData    = null;

        InputStream in = request.getInputStream();
        binaryData     = IOTools.readContent(in, SIZE_INPUT_BUFFER);

        String ipRequest = ((ServletRequest)request).getRemoteAddr();

        if (log.isTraceEnabled()) {
            log.trace("Remote address: " + ipRequest);
        }

        try {
            if (in != null) {
                in.close();
            }
        } catch (Exception e) {
            log.error("Error processing the request", e);
            return;
        }
        in = null;

        final String contentType = request.getContentType().split(";")[0];
        try {
            if (Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {

                //
                // Convert WBXML into XML
                //
                if (log.isTraceEnabled()) {
                    log.trace("Converting message from wbxml to xml");
                }
                xmlData = WBXMLTools.wbxmlToXml(binaryData);

            } else if (Constants.MIMETYPE_SYNCMLDS_XML.equals(contentType)) {
                //
                //None conversion is necessary
                //
                xmlData = new String(binaryData);
            } else {
                throw new ProtocolException( "Mime type "
                                           + contentType
                                           + " nor supported or unknown" );
            }
        } catch (Exception e) {
            log.error("Error processing the request", e);
            throw new ServletException(e);
        }

        //
        //Transform the string message into a SyncML object
        //
        if (log.isTraceEnabled()) {
            log.trace("Message to translate into the SyncML object:\n" + xmlData);
        }

        SyncML syncObj = null;
        try {
            syncObj = convert(xmlData);
        } catch (Exception e) {
            log.error("Unable to parse the request", e);
            response.setStatus(500);
            return;
        }

        if (log.isTraceEnabled()) {
            log.trace("About processing message: " + SyncMLUtil.toXML(syncObj));
        }

        //
        //Authenticate device
        //
        if (syncObj.getSyncHdr().getCred() == null) {
            if (log.isInfoEnabled()) {
                log.info("Device not authenticated! Credentials are missing.");
            }
            response.setStatus(407);
            return;
        }

        Cred   cred     = syncObj.getSyncHdr().getCred();
        String deviceId = syncObj.getSyncHdr().getSource().getLocURI();

        cred.getAuthentication().setDeviceId(deviceId);

        if (!authenticate(cred)) {
            if (log.isInfoEnabled()) {
                log.info("Device not authenticated! Invalid credentials.");
            }
            response.setStatus(401);
        } else {
            DeviceInventory devInv = (DeviceInventory)Configuration
                                                     .getConfiguration()
                                                     .getDeviceInventory();

            Sync4jDevice device = new Sync4jDevice();
            device.setDeviceId(deviceId);
            try {
                //
                //Get device without capabilities
                //
                boolean found = devInv.getDevice(device, false);
                if (!found) {
                    response.setStatus(404);
                    return;
                }

            } catch (DeviceInventoryException e) {
                log.error("Error getting the device", e);
                response.setStatus(500);
                return;
            } catch (Exception e) {
                log.error("Error getting the device", e);
                response.setStatus(500);
                return;
            }
            //
            //This flag is used to tell if device's address or(and) msisdn
            //are specificated into Alert commands
            //
            boolean notified = false;

            boolean validIpAddressReceived = false;
            boolean msisdnReceived         = false;

            if (syncObj.getSyncBody().getCommands() != null) {
                int numCmds  = syncObj.getSyncBody().getCommands().size();
                int numItems = 0;
                ArrayList items = new ArrayList();
                String  sourceURI = null;
                String  data      = null;

                for (int i = 0; i < numCmds; i++) {
                    Alert alert = (Alert) syncObj.getSyncBody().getCommands().get(i);

                    //
                    //Check message data id - must be 745 to be processed
                    //
                    if (alert.getData() == 745) {
                        //
                        //Update device with informations contains into Alert
                        //
                        if (alert.getItems() != null) {
                            items    = alert.getItems();
                            numItems = items.size();
                            for (int j = 0; j < numItems; j++) {
                                Item item = (Item)items.get(j);
                                if (item.getSource()             != null &&
                                    item.getSource().getLocURI() != null &&
                                    item.getData()               != null   ) {

                                    if (log.isTraceEnabled()) {
                                        log.trace("Item source URI : "  + item.getSource().getLocURI());
                                        log.trace("Item source DATA: "  + item.getData().getData());
                                    }

                                    sourceURI = item.getSource().getLocURI();
                                    if (sourceURI.equalsIgnoreCase(SOURCEURI_IP)) {
                                        data = item.getData().getData();
                                        //
                                        // Checks if the ip:port notified is reachable
                                        // from the server
                                        //
                                        boolean validData = checkIP(data);
                                        if (!validData) {
                                            if (log.isInfoEnabled()) {
                                                log.info("The received data (" + data + ") " +
                                                         "is not a reachable address");
                                            }
                                        } else {
                                            if (log.isInfoEnabled()) {
                                                log.info("The received data (" + data + ") " +
                                                         "is a reachable address");
                                            }
                                        }

                                        if (validData) {
                                            device.setAddress(data);
                                            //
                                            // received a valid alert with
                                            // the device ip
                                            //
                                            validIpAddressReceived = true;
                                        } else {
                                            device.setAddress("");
                                        }
                                        notified = true;
                                    }
                                    if (sourceURI.equalsIgnoreCase(SOURCEURI_MSISDN)) {
                                        device.setMsisdn(item.getData().getData());
                                        msisdnReceived = true;
                                        notified       = true;
                                    }
                                }
                            }
                        }
                    }
                }

                if (notified) {
                    try {

                        if (validIpAddressReceived || msisdnReceived) {
                            //
                            // We have received a valid alert so if the device hasn't
                            // a notification builder, we set the default one.
                            //
                            if (device.getNotificationBuilder() == null ||
                                device.getNotificationBuilder().length() == 0) {
                                device.setNotificationBuilder(defaultNotificationBuilder);
                            }
                        }

                        if (validIpAddressReceived) {
                            //
                            // We have received a valid alert with the device ip, 
                            // so we can set the device's sender to the default
                            // value of the TCPIPSender (see init method)
                            //
                            device.setNotificationSender(defaultTCPIPNotificationSender);
                        }

                        if (msisdnReceived) {
                            //
                            // We have received a msisdn. If the device
                            // hasn't a sender, we set the default wap sender
                            //
                            if (device.getNotificationSender() == null ||
                                device.getNotificationSender().length() == 0) {
                                device.setNotificationSender(defaultWAPNotificationSender);
                            }
                        }
                        devInv.setDevice(device);
                        if (log.isTraceEnabled()) {
                            log.trace("Storing " + device);
                        }
                    } catch (DeviceInventoryException e) {

                        log.error("Error processing the request", e);

                        response.setStatus(500);
                        return;
                    }

                    if (validIpAddressReceived || msisdnReceived) {
                        response.setStatus(200);
                    } else {
                        //
                        // We have received a valid Alert command (notified == true)
                        // but the ip isn't valid and there isn't the msisdn.
                        //
                        //
                        // Set status = 420 if there is an ip address includes
                        // into Alert commands that is different from the ip
                        // address of the client that sent the request
                        //
                        response.setStatus(420);
                    }
                } else {
                    if (log.isTraceEnabled()) {
                        log.trace("Bad Request");
                    }
                    response.setStatus(400);
                }
            } else {
                log.error("Message sent to AddressListener is null or " +
                          "does not contain any Alert command");
                response.setStatus(500);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Authenticates a device
     *
     * @param credentials the credentials used for authentication
     *
     * @return true if device was successfuly authenticated, false otherwise
     * @throws IOException
     */
    private boolean authenticate(Cred credentials) throws IOException {
        Officer officer = Configuration.getConfiguration().getOfficer();
        Sync4jUser user = officer.authenticateUser(credentials);
        if (user == null) {
            return false;
        }
        return true;
    }

    /**
     * Converts a SyncML message into a SyncML object
     *
     * @param msg the message to be converted
     *
     * @return the SyncML object obtained from message
     * @throws ServerException
     */
    private SyncML convert(final String msg) throws ServerException {
        try {
            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IUnmarshallingContext c = f.createUnmarshallingContext();
            return (SyncML) c.unmarshalDocument(new StringReader(msg));
        } catch (JiBXException e) {
            log.error("Error unmarshalling message", e);
            throw new ServerException(e);
        }
    }

    /**
     * Log the headers of the request
     * @param request HttpServletRequest
     */
    private void showHeaders(HttpServletRequest request) {
        Enumeration enumHeaders = request.getHeaderNames();
        StringBuffer sb = new StringBuffer("Http header: \n");
        String headerName  = null;
        String headerValue = null;
        while (enumHeaders.hasMoreElements()) {
            headerName = (String)enumHeaders.nextElement();
            headerValue = request.getHeader(headerName);
            sb.append("> ").append(headerName);
            sb.append(": ").append(headerValue).append("\n");
        }
        log.trace(sb.toString());
    }

    /**
     * Cheks if the given data contains a valid address usable to send notification
     * messages via tcp. The given data must be in the form <code>address:port</code>
     * or just <code>address</code>. The address is tested sending a fake notification
     * message without alerts.
     * @param data the address to test
     * @return boolean true if the device is reachable, false otherwise.
     */
    private boolean checkIP(String data) {

        if (data == null) {
            return false;
        }
        String address = null;
        int    port    = 745;

        int index = data.indexOf(':');
        if (index != -1) {
            address = data.substring(0, index);
            port =  Integer.parseInt(data.substring(index + 1));
        } else {
            //
            // The data doesn't contain the port. The defualt one (745) is used
            //
            address = data;
        }

        if (log.isTraceEnabled()) {
            log.trace("Checking if the ip " + address + " (port: " + port + ") is reachable" +
                      " sending a fake notification message");
        }

        byte[] fakeMessage = null;
        try {
            fakeMessage = buildFakeNotificationMessage();
        } catch (Exception ex) {
            //
            // Creating a really fake message. This was obtained calling the
            // builder offline
            //
            fakeMessage = new byte[] {
                (byte)0xF6, (byte)0xE2, (byte)0xD3, (byte)0x51, (byte)0xE1,
                (byte)0xDC, (byte)0x51, (byte)0xA5, (byte)0x86, (byte)0x96,
                (byte)0xE3, (byte)0xC8, (byte)0x06, (byte)0x53, (byte)0x60,
                (byte)0xCF, (byte)0x03, (byte)0x08, (byte)0x00, (byte)0x00,
                (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x08, (byte)0x66,
                (byte)0x75, (byte)0x6E, (byte)0x61, (byte)0x6D, (byte)0x62,
                (byte)0x6F, (byte)0x6C, (byte)0x00
            };
        }

        Socket           sock = null;
        DataOutputStream out  = null;
        BufferedReader   in   = null;

        try {

            //create a socket and connect to device
            sock = new Socket();
            SocketAddress sockAddress = new InetSocketAddress(address, port);

            sock.connect(sockAddress, 20000); // timeout = 20 seconds

            out = new DataOutputStream(sock.getOutputStream());

            //send message through socket
            out.write(fakeMessage);

            //flsuh the message in stream
            out.flush();

            //read status response
            in = new BufferedReader(
                    new InputStreamReader(sock.getInputStream())
                 );
            String response = in.readLine();

            if (log.isTraceEnabled()) {
                log.trace("Response from the client: " + response);
            }

            return true;

        } catch (Exception e) {

            if (log.isInfoEnabled()) {
                log.info("Error sending fake notification message to " +
                         address +
                         " (port: " + port + "): " + e.toString());
            }

        } finally {

            //close connection to device
            try {
                if (out != null) {
                    out.close();
                }
                if (in != null) {
                    in.close();
                }
                if (sock != null) {
                    if (!sock.isClosed()) {
                        sock.close();
                    }
                }
            } catch (IOException e) {
                if (log.isTraceEnabled()) {
                    log.trace("Error closing connection (" + e.toString() + ")");
                }
            }
        }

        return false;
    }

    /**
     * Creates a fake notifcation message using a <code>DSNotificationMessageBuilderImpl</code>.
     * The fake message doesn't contains valid information and alert commands but
     * it is a well-format notification message.
     *
     * @return byte[] the fake message
     */
    private static byte[] buildFakeNotificationMessage() throws Exception {
        byte[]  fakeServerNonce   = new byte[] {0};
        String  fakeServerId      = "funambol";
        String  fakeServerPw      = "funambol";
        String  fakeDeviceId      = "device";
        int     fakeUIMode        = 0;
        String  fakeDeviceAddress = "";
        int     fakeSessionId     = 0;
        Alert[] fakeAlerts        = new Alert[0];

        DSNotificationMessageBuilder messageBuilder = new DSNotificationMessageBuilderImpl();



        Message fakeMessage = messageBuilder.buildMessage(fakeSessionId,
                                                          fakeDeviceId,
                                                          fakeDeviceAddress,
                                                          fakeServerId,
                                                          fakeServerPw,
                                                          fakeServerNonce,
                                                          fakeAlerts,
                                                          fakeUIMode,
                                                          1.2f);

        return fakeMessage.getMessageContent();
    }

}
